@@ -1,21 +1,25 @@
+import errno
 import hashlib
 import json
 import os
+import re
 import shlex
 import shutil
 import subprocess
 import tempfile
 import warnings
 from collections import Counter
-from datetime import datetime, timedelta
+from datetime import datetime
 from functools import lru_cache
+from itertools import groupby
+from typing import Iterable, Set
 
-from luigi.target import Target
 from b2luigi.basf2_helper.utils import get_basf2_git_hash
 from b2luigi.batch.processes import BatchProcess, JobStatus
 from b2luigi.core.settings import get_setting
 from b2luigi.core.utils import flatten_to_dict, get_log_file_dir, get_task_file_dir
 from jinja2 import Template
+from luigi.target import Target
 
 
 class Gbasf2Process(BatchProcess):
@@ -25,10 +29,13 @@ class Gbasf2Process(BatchProcess):
     Features
         - **gbasf2 project submission**
 
-          The gbasf2 batch process takes the basf2 path returned by the ``create_path()``
-          method of the task, saves it to the disk and creates a wrapper steering file that
-          executes the saved path. It then sends both to the LCG via
-          the BelleII-specific Dirac-wrapper gbasf2.
+          The gbasf2 batch process takes the basf2 path returned by the
+          ``create_path()`` method of the task, saves it into a pickle file to
+          the disk and creates a wrapper steering file that executes the saved
+          path. Any basf2 variable aliases added in the ``create_path()`` method
+          are also stored in the pickle file. It then sends both the pickle file
+          and the steering file wrapper to the grid via the BelleII-specific
+          Dirac-wrapper gbasf2.
 
         - **Project status monitoring**
 
@@ -63,9 +70,10 @@ class Gbasf2Process(BatchProcess):
           inhereting from ``Basf2PathTask`` or other tasks with a
           ``create_path()`` method that returns a basf2 path.
 
-        - It can be used **only for pickable basf2 paths**, as it stores
+        - It can be used **only for pickable basf2 paths**, with only some limited global basf2 state
+          saved (currently aliases and global tags). The batch process stores
           the path created by ``create_path`` in a python pickle file and runs that on the grid.
-          Therefore, **basf2 variable aliases and python basf2 modules are not yet supported**. 
+          Therefore, **python basf2 modules are not yet supported**.
           To see if the path produced by a steering file is pickable, you can try to dump it with
           ``basf2 --dump-path`` and execute it again with ``basf2 --execute-path``.
 
@@ -75,6 +83,7 @@ class Gbasf2Process(BatchProcess):
           ``B_ntuple.root``), but a collection of root files, one for each file in
           the input data set, in a directory with the base name of the root
           files, e.g.::
+
             <task output directory>
                         ├── B_ntuple.root
                         │   └── B_ntuple_0.root
@@ -82,7 +91,7 @@ class Gbasf2Process(BatchProcess):
                         │   └── ...
                         ├── D_ntuple.root
                         │   └── D_ntuple_0.root
-                        │   └── ... 
+                        │   └── ...
 
 
 
@@ -142,6 +151,9 @@ class Gbasf2Process(BatchProcess):
           as they don't need the output locally but take the grid datasets as input. Also useful when you just want
           to produce data on the grid for other people to use.
 
+        - ``gbasf2_download_logs``: Whether to automatically download the log output of gbasf2 projects when the
+          task succeeds or fails. Having the logs is important for reproducibility, but k
+
         The following optional settings correspond to the equally named ``gbasf`` command line options
         (without the ``gbasf_`` prefix) that you can set to customize your gbasf2 project:
 
@@ -239,24 +251,7 @@ class Gbasf2Process(BatchProcess):
         to ``gb2_job_status``, and return an overall project status, e.g. when
         all jobs are done, return ``JobStatus.successful`` to show that the
         gbasf2 project succeeded.
-
-        The status of each individual job can be one of::
-
-            [Submitting, Submitted, Received, Checking, Staging, Waiting, Matched, Rescheduled,
-             Running, Stalled, Completing, Done, Completed, Failed, Deleted, Killed]
-
-        (Taken from  https://github.com/DIRACGrid/DIRAC/blob/rel-v7r1/WorkloadManagementSystem/Client/JobStatus.py)
-
         """
-        # If project is does not exist on grid yet, so can't query for gbasf2 project status
-        if not check_project_exists(self.gbasf2_project_name, dirac_user=self.dirac_user):
-            raise RuntimeError(
-                f"\nCould not find any jobs for project {self.gbasf2_project_name} on the grid.\n" +
-                "Probably there was an error during the project submission when running the gbasf2 command.\n" +
-                "Try if you can run the gbasf2 command used manually in a terminal with gbasf2 set up:\n" +
-                " ".join(self._build_gbasf2_submit_command())
-            )
-
         job_status_dict = get_gbasf2_project_job_status_dict(self.gbasf2_project_name, dirac_user=self.dirac_user)
         n_jobs_by_status = Counter()
         for _, job_info in job_status_dict.items():
@@ -292,8 +287,15 @@ class Gbasf2Process(BatchProcess):
         if n_done == n_jobs:
             # download dataset only the first time that we return JobStatus.successful
             if not self._project_had_been_successful:
-                self._on_first_success_action()
-                self._project_had_been_successful = True
+                try:
+                    self._on_first_success_action()
+                    self._project_had_been_successful = True
+                # RuntimeError might occur when download of output dataset was not complete. This is
+                # frequent, so we want to catch that error and just marking the job as failed
+                except RuntimeError as err:
+                    warnings.warn(err, RuntimeWarning)
+                    return JobStatus.aborted
+
             return JobStatus.successful
 
         raise RuntimeError("Could not determine JobStatus")
@@ -302,7 +304,8 @@ class Gbasf2Process(BatchProcess):
         """
         Things to do after all jobs in the project had been successful, e.g. downloading the dataset and logs
         """
-        self._download_logs()
+        if get_setting("gbasf2_download_logs", default=True, task=self.task):
+            self._download_logs()
         if get_setting("gbasf2_download_dataset", default=True, task=self.task):
             self._download_dataset()
 
@@ -315,7 +318,8 @@ class Gbasf2Process(BatchProcess):
                            if job_info["Status"] == "Failed"}
         n_failed = len(failed_job_dict)
         print(f"{n_failed} failed jobs:\n{failed_job_dict}")
-        self._download_logs()
+        if get_setting("gbasf2_download_logs", default=True, task=self.task):
+            self._download_logs()
 
     def _reschedule_failed_jobs(self):
         """
@@ -327,7 +331,10 @@ class Gbasf2Process(BatchProcess):
         for job_id, job_info in job_status_dict.items():
             if job_info["Status"] == "Failed":
                 if self.n_retries_by_job[job_id] >= self.max_retries:
-                    warnings.warn(f"Reached maximum number of rescheduling tries ({self.max_retries}) for job {job_id}.")
+                    warnings.warn(
+                        f"Reached maximum number of rescheduling tries ({self.max_retries}) for job {job_id}.",
+                        RuntimeWarning
+                    )
                     return False
                 self._reschedule_job(job_id)
                 self.n_retries_by_job[job_id] += 1
@@ -350,7 +357,7 @@ class Gbasf2Process(BatchProcess):
         Submit new gbasf2 project to grid
         """
         if check_project_exists(self.gbasf2_project_name, self.dirac_user):
-            warnings.warn(
+            print(
                 f"\nProject with name {self.gbasf2_project_name} already exists on grid, "
                 "therefore not submitting new project. If you want to submit a new project, "
                 "change the project name."
@@ -395,8 +402,9 @@ class Gbasf2Process(BatchProcess):
         """
         gbasf2_release = get_setting("gbasf2_release", default=get_basf2_git_hash(), task=self.task)
         gbasf2_additional_files = get_setting("gbasf2_additional_files", default=[], task=self.task)
-        assert not isinstance(gbasf2_additional_files, str), "gbasf2_additional_files should be a list or tuple, not a string."
-        gbasf2_input_sandbox_files = [os.path.basename(self.pickle_file_path)] + gbasf2_additional_files
+        if not isinstance(gbasf2_additional_files, Iterable) or isinstance(gbasf2_additional_files, str):
+            raise ValueError("``gbasf2_additional_files`` is not an iterable or strings.")
+        gbasf2_input_sandbox_files = [os.path.basename(self.pickle_file_path)] + list(gbasf2_additional_files)
         gbasf2_command_str = (f"gbasf2 {self.wrapper_file_path} -f {' '.join(gbasf2_input_sandbox_files)} " +
                               f"-p {self.gbasf2_project_name} -s {gbasf2_release} ")
 
@@ -428,7 +436,8 @@ class Gbasf2Process(BatchProcess):
         # gbasf2 job priority
         priority = get_setting("gbasf2_priority", default=False, task=self.task)
         if priority is not False:
-            assert 0 <= priority <= 10, "Priority should be integer between 0 and 10."
+            if not (0 <= priority <= 10):
+                raise ValueError("Priority should be integer between 0 and 10.")
             gbasf2_command_str += f" --priority {priority} "
 
         # gbasf2 job type (e.g. User, Production, ...)
@@ -443,7 +452,7 @@ class Gbasf2Process(BatchProcess):
 
         # optional string of additional parameters to append to gbasf2 command
         gbasf2_additional_params = get_setting("gbasf2_additional_params", default=False, task=self.task)
-        if basf2opt is not False:
+        if gbasf2_additional_params is not False:
             gbasf2_command_str += f" {gbasf2_additional_params} "
 
         gbasf2_command = shlex.split(gbasf2_command_str)
@@ -452,6 +461,8 @@ class Gbasf2Process(BatchProcess):
     def _write_path_to_file(self):
         """
         Serialize and save the ``basf2.Path`` returned by ``self.task.create_path()`` to a python pickle file.
+
+        Also serializes any basf2 variable aliases from the current variable manager instance.
         """
         try:
             basf2_path = self.task.create_path()
@@ -460,8 +471,8 @@ class Gbasf2Process(BatchProcess):
                 "Gbasf2 batch process can only used with tasks that generate basf2 paths with "
                 "a ``create_path()`` method, e.g. are an instance of ``Basf2PathTask``."
             ) from err
-        from basf2.pickle_path import write_path_to_file
-        write_path_to_file(basf2_path, self.pickle_file_path)
+        from b2luigi.batch.processes.gbasf2_utils.pickle_utils import write_path_and_state_to_file
+        write_path_and_state_to_file(basf2_path, self.pickle_file_path)
 
     def _create_wrapper_steering_file(self):
         """
@@ -480,6 +491,72 @@ class Gbasf2Process(BatchProcess):
             # write the template with the replacements to a new file which should be sent to the grid
             steering_file_stream.dump(self.wrapper_file_path)
 
+    def _get_gbasf2_dataset_query(self, output_file_name: str) -> str:
+        """
+        Helper method that returns the gbasf2 query string with the correct wildcard pattern
+        to get the subset of all files for ``output_file_name`` from the grid project associated with this task,
+        either, e.g. via the ``gb2_ds_list`` or ``gb2_ds_get`` commands.
+
+        Args:
+            output_file_name: Output file name, must be a root file, e.g. ``ntuple.root``.
+                Usually defined by the user via :py:func:`b2luigi.Task.add_to_output` in
+                the :py:func:`b2luigi.Task.output` method.
+        """
+        if output_file_name != os.path.basename(output_file_name):
+            raise ValueError(
+                f"For grid projects, the output file name must not be a basename, not a path, but is \"{output_file_name}\""
+            )
+        output_file_stem, output_file_ext = os.path.splitext(output_file_name)
+        if output_file_ext != ".root":
+            raise ValueError(
+                f"Output file name \"{output_file_name}\" does not end with \".root\", "
+                "but gbasf2 batch only supports root outputs"
+            )
+        dataset_query_string = \
+            f"/belle/user/{self.dirac_user}/{self.gbasf2_project_name}/sub00/{output_file_stem}_*{output_file_ext}"
+        return dataset_query_string
+
+    def _local_gb2_dataset_is_complete(self, output_file_name: str, check_temp_dir: bool = False, verbose: bool = False) -> bool:
+        """
+        Helper method that returns ``True`` if the download of the gbasf2
+        dataset for the output ``output_file_name`` is complete.
+
+        Args:
+            output_file_name: Output file name, must be a root file, e.g. ``ntuple.root``.
+                Usually defined by the user via :py:func:`b2luigi.Task.add_to_output` in
+                the :py:func:`b2luigi.Task.output` method.
+            check_temp_dir: Instead of checking the final output path, check whether the download into the
+                temporary ("partial") directory is complete. This function is usually called with this
+                argument set to ``True``, to check whether the dataset can be moved to its final output path.
+            verbose: If true, print filenames in datasets.
+        """
+        # first get the local set of files in the dataset for `output_file_name`
+        task_output_dict = flatten_to_dict(self.task.output())
+        output_target = task_output_dict[output_file_name]
+        output_dir_path = output_target.path
+        if check_temp_dir:
+            output_dir_path = os.path.join(output_dir_path + ".partial", self.gbasf2_project_name, 'sub00')
+        if not os.path.isdir(output_dir_path):
+            raise FileNotFoundError(errno.ENOTDIR, os.strerror(errno.ENOTDIR), output_dir_path)
+        downloaded_dataset_basenames = os.listdir(output_dir_path)
+
+        # not get the remote set of grid file names for the gbasf2 project output matching output_file_name
+        ds_query_string = self._get_gbasf2_dataset_query(output_file_name)
+        ds_list_command = shlex.split(f"gb2_ds_list {ds_query_string}")
+        output_dataset_grid_filepaths = run_with_gbasf2(ds_list_command, capture_output=True).stdout.splitlines()
+        output_dataset_basenames = [os.path.basename(grid_path) for grid_path in output_dataset_grid_filepaths]
+        # remove duplicate LFNs that gb2_ds_list returns for outputs from rescheduled jobs
+        output_dataset_basenames = get_unique_lfns(output_dataset_basenames)
+        # check if local and remote datasets are equal
+        if set(output_dataset_basenames) == set(downloaded_dataset_basenames):
+            return True
+        if verbose:
+            print(
+                "\nDownloaded files:\n{}".format("\n".join(downloaded_dataset_basenames)) +
+                "\nFiles on the grid:\n{}".format("\n".join(output_dataset_basenames))
+            )
+        return False
+
     def _download_dataset(self):
         """
         Download the task outputs from the gbasf2 project dataset.
@@ -494,47 +571,39 @@ class Gbasf2Process(BatchProcess):
             raise RuntimeError(f"Not dataset to download under project name {self.gbasf2_project_name}")
         task_output_dict = flatten_to_dict(self.task.output())
         for output_file_name, output_target in task_output_dict.items():
+            dataset_query_string = self._get_gbasf2_dataset_query(output_file_name)
             output_dir_path = output_target.path
-            assert output_file_name == os.path.basename(output_file_name)  # not sure I need this
-            output_file_stem, output_file_ext = os.path.splitext(output_file_name)
-            assert output_file_ext == ".root", "gbasf2 batch only supports root outputs"
-
-            # Get list of files that we want to download from the grid via ``gb2_ds_list`` so that we can
-            # then compare this list with the results of the download to see if it was successful
-            dataset_query_string = f"--user {self.dirac_user} {self.gbasf2_project_name}/{output_file_stem}_*{output_file_ext}"
-            ds_list_command = shlex.split(f"gb2_ds_list {dataset_query_string}")
-            output_dataset_grid_filepaths = run_with_gbasf2(ds_list_command, capture_output=True).stdout.splitlines()
-            output_dataset_basenames = {os.path.basename(grid_path) for grid_path in output_dataset_grid_filepaths}
             # check if dataset had been already downloaded and if so, skip downloading
-            if os.path.isdir(output_dir_path) and os.listdir(output_dir_path) == output_dataset_basenames:
+            if os.path.isdir(output_dir_path) and self._local_gb2_dataset_is_complete(output_file_name):
                 print(f"Dataset already exists in {output_dir_path}, skipping download.")
-                return
+                continue
 
             # To prevent from task being accidentally marked as complete when the gbasf2 dataset download failed,
-            # we create a temporary directory in the parent of ``output_dir_path`` and first download the dataset there.
-            # The download command will download it into a subdirectory with the same name as the project.
+            # we create a temporary directory and first download the dataset there.
             # If the download had been successful and the local files are identical to the list of files on the grid,
-            # we move the downloaded dataset to the location specified by ``output_dir_path``
-            output_dir_parent = os.path.dirname(output_dir_path)
-            os.makedirs(output_dir_parent, exist_ok=True)
-            with tempfile.TemporaryDirectory(dir=output_dir_parent) as tmpdir_path:
-                ds_get_command = shlex.split(f"gb2_ds_get --force {dataset_query_string}")
-                print("Downloading dataset with command ", " ".join(ds_get_command))
-                stdout = run_with_gbasf2(ds_get_command, cwd=tmpdir_path, capture_output=True).stdout
-                print(stdout)
-                if "No file found" in stdout:
-                    raise RuntimeError(f"No output data for gbasf2 project {self.gbasf2_project_name} found.")
-                tmp_output_dir = os.path.join(tmpdir_path, self.gbasf2_project_name)
-                downloaded_dataset_basenames = set(os.listdir(tmp_output_dir))
-                if output_dataset_basenames == downloaded_dataset_basenames:
-                    print(f"Download of {self.gbasf2_project_name} files successful.\n"
-                          f"Moving output files to directory: {output_dir_path}")
-                    if os.path.exists(output_dir_path):
-                        shutil.rmtree(output_dir_path)
-                    shutil.move(src=tmp_output_dir, dst=output_dir_path)
-                else:
-                    raise RuntimeError(f"The downloaded of files in {tmp_output_dir} is not equal to the "
-                                       f"dataset files for the grid project {self.gbasf2_project_name}")
+            # we move the downloaded dataset to the location specified by ``output_dir_path``.
+            tmp_output_dir_path = f"{output_dir_path}.partial"
+            os.makedirs(tmp_output_dir_path, exist_ok=True)
+
+            ds_get_command = shlex.split(f"gb2_ds_get --force {dataset_query_string}")
+            print("Downloading dataset with command ", " ".join(ds_get_command))
+            stdout = run_with_gbasf2(ds_get_command, cwd=tmp_output_dir_path, capture_output=True).stdout
+            print(stdout)
+            if "No file found" in stdout:
+                raise RuntimeError(f"No output data for gbasf2 project {self.gbasf2_project_name} found.")
+            tmp_output_dir = os.path.join(tmp_output_dir_path, self.gbasf2_project_name, 'sub00')
+            if not self._local_gb2_dataset_is_complete(output_file_name, check_temp_dir=True, verbose=True):
+                raise RuntimeError(
+                    f"Download incomplete. The downloaded set of files in {tmp_output_dir} is not equal to the " +
+                    f"list of dataset files on the grid for project {self.gbasf2_project_name}.",
+                )
+
+            print(f"Download of {self.gbasf2_project_name} files successful.\n"
+                  f"Moving output files to directory: {output_dir_path}")
+            if os.path.exists(output_dir_path):
+                shutil.rmtree(output_dir_path)
+            shutil.move(src=tmp_output_dir, dst=output_dir_path)
+            shutil.rmtree(tmp_output_dir_path)
 
     def _download_logs(self):
         """
@@ -578,6 +647,7 @@ class Gbasf2GridProjectTarget(Target):
     Target exists if an output dataset for the project exists on the grid and is
     not being written to, i.e. all jobs that produced the dataset are done.
     """
+
     def __init__(self, project_name, dirac_user=None):
         """
         :param project_name: Name of the gbasf2 grid project that produced the
@@ -613,16 +683,19 @@ def check_dataset_exists_on_grid(gbasf2_project_name, dirac_user=None):
         return False
     output_lines_are_paths = all(os.path.abspath(line) for line in output_dataset_str.strip().splitlines())
     if not output_lines_are_paths:
-        warnings.warn("The output of ``{' '.join(ds_list_command)}`` contains lines that are not grid paths:\n" +
-                      output_dataset_str)
+        warnings.warn(
+            "The output of ``{' '.join(ds_list_command)}`` contains lines that are not grid paths:\n" +
+            output_dataset_str,
+            category=RuntimeWarning
+        )
         return False
     return True
 
 
 def get_gbasf2_project_job_status_dict(gbasf2_project_name, dirac_user=None):
     """
-    Returns a dictionary for all jobs in the project with a structure like the following,
-    which I have taken and adapted from an example output::
+    Returns a dictionary for all jobs in the project with a structure like the
+    following, which I have taken and adapted from an example output::
 
         {
             "<JobID>": {
@@ -639,19 +712,24 @@ def get_gbasf2_project_job_status_dict(gbasf2_project_name, dirac_user=None):
         ...
         }
 
-    For that purpose, the script in ``gbasf2_job_status.py`` is called.
-    That script directly interfaces with Dirac via its API, but it only works with the
-    gbasf2 environment and python2, which is why it is called as a subprocess.
-    The job status dictionary is passed to this function via json.
+    For that purpose, the script in ``gbasf2_job_status.py`` is called.  That
+    script directly interfaces with Dirac via its API, but it only works with
+    the gbasf2 environment and python2, which is why it is called as a
+    subprocess.  The job status dictionary is passed to this function via json.
     """
     if dirac_user is None:
         dirac_user = get_dirac_user()
-    assert check_project_exists(gbasf2_project_name, dirac_user), \
-        f"Project {gbasf2_project_name} doest not exist yet"
     job_status_script_path = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                           "gbasf2_utils/gbasf2_job_status.py")
-    job_status_command = shlex.split(f"python2 {job_status_script_path} -p {gbasf2_project_name} --user {dirac_user}")
-    job_status_json_string = run_with_gbasf2(job_status_command, capture_output=True).stdout
+    job_status_command = shlex.split(f"{job_status_script_path} -p {gbasf2_project_name} --user {dirac_user}")
+    proc = run_with_gbasf2(job_status_command, capture_output=True, check=False)
+    # FIXME: use enum or similar to define my own return codes
+    if proc.returncode == 3:  # return code 3 means project does not exist yet
+        raise RuntimeError(
+            f"\nCould not find any jobs for project {gbasf2_project_name} on the grid.\n" +
+            "Probably there was an error during the project submission when running the gbasf2 command.\n"
+        )
+    job_status_json_string = proc.stdout
     job_status_dict = json.loads(job_status_json_string)
     return job_status_dict
 
@@ -660,22 +738,20 @@ def check_project_exists(gbasf2_project_name, dirac_user=None):
     """
     Check if we can find the gbasf2 project on the grid with ``gb2_job_status``.
     """
-    if dirac_user is None:
-        dirac_user = get_dirac_user()
-    command = shlex.split(f"gb2_job_status -p {gbasf2_project_name} --user {dirac_user}")
-    output = run_with_gbasf2(command, capture_output=True).stdout
-    if output.strip() == "0 jobs are selected.":
+    try:
+        return bool(get_gbasf2_project_job_status_dict(gbasf2_project_name, dirac_user))
+    except RuntimeError:
         return False
-    if "--- Summary of Selected Jobs ---" in output:
-        return True
-    raise RuntimeError("Output of gb2_job_status did not contain expected strings,"
-                       " could not determine if project exists")
 
 
-def run_with_gbasf2(cmd, *args, check=True, encoding="utf-8", capture_output=False, **kwargs):
+def run_with_gbasf2(
+        cmd, *args, ensure_proxy_initialized=True, check=True, encoding="utf-8", capture_output=False, **kwargs
+):
     """
     Call ``cmd`` in a subprocess with the gbasf2 environment.
 
+    :param ensure_proxy_initialized: If this is True, check if the dirac proxy is initalized and alive and if not,
+                                     initialize it.
     :param check: Whether to raise a ``CalledProcessError`` when the command returns with an error code.
                   The default value ``True`` is the same as in ``subprocess.check_call()`` and different as in the
                   normal ``run_with_gbasf2()`` command.
@@ -695,6 +771,8 @@ def run_with_gbasf2(cmd, *args, check=True, encoding="utf-8", capture_output=Fal
         kwargs['stdout'] = subprocess.PIPE
         kwargs['stderr'] = subprocess.PIPE
     gbasf2_env = get_gbasf2_env()
+    if ensure_proxy_initialized:
+        setup_dirac_proxy()
     proc = subprocess.run(cmd, *args, check=check, encoding=encoding, env=gbasf2_env, **kwargs)
     return proc
 
@@ -733,7 +811,7 @@ def get_gbasf2_env(gbasf2_install_directory=None):
 
 def get_dirac_user():
     """Get dirac user name"""
-    proxy_info_str = setup_dirac_proxy()
+    proxy_info_str = run_with_gbasf2(["gb2_proxy_info"], capture_output=True).stdout
     for line in proxy_info_str.splitlines():
         if line.startswith("username"):
             dirac_user = line.split(":", 1)[1].strip()
@@ -743,29 +821,20 @@ def get_dirac_user():
 
 def setup_dirac_proxy():
     """
-    Runs ``gb2_proxy_init -g belle`` if necessary and returns
-    ``gb2_proxy_info`` output
-
-    ``gb2_proxy_init`` has to be run only once every 24 hours for a
-    terminal.  So we first check with ``gb2_proxy_info`` and if an old proxy
-    is still valid, don't re-run the command to prevent unnecessary
-    certificate password prompts
+    Runs ``gb2_proxy_init -g belle`` if there's no active dirac proxy. If there is, do nothing.
     """
-    # Get time that the proxy is still valid from the gb2_proxy_info output line "timeleft".
-    # If no proxy had been initialized, the output will not contain the "timeleft" string.
-    # Alternatively, if the proxy time ran out, the timeleft value will be 00:00:00
-    proxy_info_str = run_with_gbasf2(["gb2_proxy_info"], capture_output=True).stdout
-    for line in proxy_info_str.splitlines():
-        if line.startswith("timeleft"):
-            timeleft_str = line.split(":", 1)[1].strip()
-            timeleft = datetime.strptime(timeleft_str, '%H:%M:%S')
-            timeleft_delta = timedelta(hours=timeleft.hour, minutes=timeleft.minute, seconds=timeleft.second)
-            if timeleft_delta.total_seconds() > 0:
-                return proxy_info_str
+    check_proxy_initizalized_script_path = os.path.join(
+        os.path.dirname(os.path.realpath(__file__)),
+        "gbasf2_utils/check_if_dirac_proxy_is_initialized.py"
+    )
+    # first run script to check if proxy is already alive or needs to be initalized
+    # setting ``initalize_proxy=False`` is vital here, otherwise we get an infinite loop
+    proc = run_with_gbasf2([check_proxy_initizalized_script_path], ensure_proxy_initialized=False, check=False)
+    # if returncode of the script is 0, that means that proxy is already alive
+    if not proc.returncode:
+        return
     # initiallize proxy
-    run_with_gbasf2(shlex.split("gb2_proxy_init -g belle"))
-    new_proxy_info_str = run_with_gbasf2(["gb2_proxy_info"], capture_output=True).stdout
-    return new_proxy_info_str
+    run_with_gbasf2(shlex.split("gb2_proxy_init -g belle"), ensure_proxy_initialized=False)
 
 
 def get_unique_project_name(task):
@@ -791,10 +860,69 @@ def get_unique_project_name(task):
     task_id_hash = hashlib.md5(task.task_id.encode()).hexdigest()[0:10]
     gbasf2_project_name = gbasf2_project_name_prefix + task_id_hash
     max_project_name_length = 32
-    assert len(gbasf2_project_name) <= max_project_name_length,\
-        f"Maximum length of project name should be {max_project_name_length}, " + \
-        f"but has {len(gbasf2_project_name)} chars." + \
-        f"Please choose a gbasf2_project_name_prefix of less than {max_project_name_length - len(task_id_hash)} characters," + \
-        f" since the unique task id hash takes {len(task_id_hash)} characters."
-    assert gbasf2_project_name.isalnum(), "Only alphanumeric project names are officially supported by gbasf2"
+    if len(gbasf2_project_name) > max_project_name_length:
+        raise ValueError(
+            f"Maximum length of project name should be {max_project_name_length}, "
+            f"but has {len(gbasf2_project_name)} chars."
+            "Please choose a gbasf2_project_name_prefix of less than "
+            f"{max_project_name_length - len(task_id_hash)} characters,"
+            f" since the unique task id hash takes {len(task_id_hash)} characters."
+        )
+    # Only alphanumeric characters (letters, numbers and `_`, `-`) are supported by gbasf2
+    valid_project_name_regex_str = r"^[a-zA-Z0-9_-]*$"
+    if not re.match(valid_project_name_regex_str, gbasf2_project_name):
+        raise ValueError(
+            f"Project name \"{gbasf2_project_name}\" is invalid. "
+            "Only alphanumeric project names are officially supported by gbasf2."
+        )
     return gbasf2_project_name
+
+
+def lfn_follows_gb2v5_convention(lfn: str) -> bool:
+    """
+    Check if the LFN follows the convention of gbasf2 release 5, i.e.
+       <name>_<gbasf2param>_<jobID>_<rescheduleNum>.root
+
+    Args:
+        lfn: Logical file name, a file path on the grid
+    """
+    # adapted the logic from the BelleDirac function ``findDuplicatedJobID()``
+    if len(lfn.split('_')) < 2 or 'job' not in lfn.split('_')[-2]:
+        return False
+    return True
+
+
+def _get_lfn_upto_reschedule_number(lfn: str) -> str:
+    """
+    Get a string of the gbasf2 v5 LFN upto the reschule number.
+
+    E.g. if the LFN is ``<name>_<gbasf2param>_<jobID>_<rescheduleNum>.root``
+    return ````<name>_<gbasf2param>_<jobID>``.
+    """
+    if not lfn_follows_gb2v5_convention(lfn):
+        raise ValueError(
+            "LFN does does not conform to the new gbasf2 v5 style, "
+            "thus getting the substring upto the reschedule number does not make sense"
+        )
+    return "_".join(lfn.split("_")[:-1])
+
+
+def get_unique_lfns(lfns: Iterable[str]) -> Set[str]:
+    """
+    From list of gbasf2 LFNs which include duplicate outputs for rescheduled
+    jobs return filtered list which only include the LFNs for the jobs with the
+    highest reschedule number.
+
+    Gbasf2 v5 has outputs of the form
+    ``<name>_<gbasf2param>_<jobID>_<rescheduleNum>.root``.  When using
+    ``gb2_ds_list``, we see duplicates LFNs where all parts are the same accept
+    the ``rescheduleNum``.  This function returns only those with the highest number.
+    """
+    # if dataset does not follow gbasf2 v5 convention, assume it was produced
+    # with old release and does not contain duplicates
+    if not all(lfn_follows_gb2v5_convention(lfn) for lfn in lfns):
+        return list(lfns)
+    # if it is of the gbasf v5 form, group the outputs by the substring upto the
+    # reschedule number and return list of maximums of each group
+    lfns = sorted(lfns, key=_get_lfn_upto_reschedule_number)
+    return {max(lfn_group) for _, lfn_group in groupby(lfns, key=_get_lfn_upto_reschedule_number)}
